/* alloc.h */
/* squelette de definition des macros d'allocation */

#ifndef _alloc_
#define _alloc_

#include "config.h"
#include "mlvalues.h"

/* PUSH_ROOTS et POP_ROOTS servent a gerer la liste des racines C,
 * c'est-a dire l'ensemble des variables C qui pointent dans le tas.
 * Il est indispensable que toutes les "values" du code C se trouvent dedans
 * quand on appelle le GC.
 * Utilisation :
 * A la fin des declarations de variables locales de C, ajouter :
 * PUSH_ROOTS (nom_de_variable, taille);
 * La taille est le nombre de racines declarees. On les accede par :
 * nom_de_variable [0] ... nom_de_variable [taille - 1]
 * Le nom_de_variable NE doit PAS etre "_".
 * Juste avant la fin de la fonction, ajouter : POP_ROOTS;
 */

extern value *C_roots_head;

#define PUSH_ROOTS(name, size) \
   value name [(size) + 2]; \
   { long _; for (_ = 0; _ < (size); name [_++] = MLINT (0)); } \
   name [(size)] = (value) (size); \
   name [(size) + 1] = (value) C_roots_head; \
   C_roots_head = &(name [(size)]);

#define POP_ROOTS() {C_roots_head = (value *) C_roots_head [1]; }

extern address yp, young_end;

/* Si on a des racines hors des zones allouees par PUSH_ROOTS, on doit les
 * y mettre avant le GC et les reprendre apres le GC.
 * Pour ce faire, les macros d'allocation appellent les macros
 * SETUP_FOR_GC et RESTORE_AFTER_GC.
 * SETUP_FOR_GC doit ranger les variables locales qui sont des pointeurs dans
 * le tas, et RESTORE_AGTER_GC doit leur rendre leur valeur, qui a pu etre
 * modifiee par le GC.
 * Inutile de chercher a comprendre, c'est pour les cas tres particuliers.
 */

#define SETUP_FOR_GC {}
#define RESTORE_AFTER_GC {}

/* Pour allouer un petit bloc (size < MAX_YOUNG_SIZE).
   Prend en arguments le header et la taille, met le resultat dans "result".
*/

#define SMALL_ALLOC(result, hdr, size)                                        \
/* value result; header hdr; blocksize size; */                               \
{                                                                             \
  result = OBJECT (yp);                                                       \
  yp += (size) + 1;                                                           \
  if (yp > young_end){                                                        \
    value result_val;                                                         \
    yp -= (size) + 1;                                                         \
    result = MLINT (0);		/* peut etre une racine... */                 \
    SETUP_FOR_GC;                                                             \
    result_val = OBJECT (call_minor_gc ((size) + 1));                         \
    RESTORE_AFTER_GC;                                                         \
    result = result_val;                                                      \
  }                                                                           \
  HEADER (result) = (hdr);                                                    \
}

/* pour allouer un bloc dans le tas majeur directement. */

#define BIG_ALLOC(result, hdr, size)					      \
/* value result; header hdr; blocksize size; */		         	      \
{ value result_val;							      \
  SETUP_FOR_GC;								      \
  result_val = big_alloc ((hdr), (size));				      \
  RESTORE_AFTER_GC;							      \
  (result) = result_val;						      \
}

/* pour allouer un bloc de taille quelconque. */

#define ALLOC(result, hdr, size)					      \
/* value result; header hdr; blocksize size; */   			      \
{									      \
  if ((size) < MAX_YOUNG_SIZE){						      \
    SMALL_ALLOC ((result), (hdr), (size));				      \
  }else{								      \
    BIG_ALLOC ((result), (hdr), (size));				      \
  }									      \
}

/* pour allouer un bloc en initialisant ses champs */

#define ALLOC_INIT(result, hdr, size, init)                                   \
/* value result; header hdr; blocksize size; value init;*/                    \
{                                                                             \
  if ((size) < MAX_YOUNG_SIZE){                                               \
    value * to;                                                               \
    blocksize n;                                                              \
    SMALL_ALLOC ((result), (hdr), (size));                                    \
    for (n = (size), to = &FIELD((result),0); n; n--) *to++ = (init);         \
  }else{                                                                      \
    value result_val;                                                         \
    value init_val = (init);                                                  \
    SETUP_FOR_GC;                                                             \
    result_val = big_alloc_init ((hdr), (size), init_val);                    \
    RESTORE_AFTER_GC;                                                         \
    (result) = result_val;                                                    \
  }                                                                           \
}

/* pour allouer une chaine CAML dans une fonction C */

#define ALLOC_STRING(result, size)                                            \
{ blocksize size_in_words = ((size) + 5) >> 2;                                \
  ALLOC((result), MAKE_HEADER(size_in_words, TAG_STRING), size_in_words);     \
  BYTE ((result), (size)) = 0;                                                \
  BYTE ((result), (size_in_words << 2) - 1) = (size_in_words << 2) - (size);  \
}

#ifdef ANSI
extern value * call_minor_gc(blocksize);
extern value big_alloc(header, blocksize);
extern value big_alloc_init(header, blocksize, value);
extern value big_alloc_init_up(header, blocksize);
extern value big_alloc_init_down(header, blocksize);
extern value alloc_tuple(blocksize);
extern value alloc_string(char *);
extern value alloc_string_array(int, char **);
extern value alloc_double(double);
#else
extern value * call_minor_gc();
extern value big_alloc();
extern value big_alloc_init();
extern value big_alloc_up();
extern value big_alloc_down();
extern value alloc_tuple();
extern value alloc_string();
extern value alloc_string_array();
extern value alloc_double();
#endif

#endif /* _alloc_ */
